unit lexlib;
(*

  3.
  2001/10/24 Modified by Alexandre Alapetite <alapetite@fr.st> & Sylvain Piry
  to handle all text formats
  PC (#13#10), PC inversed (#10#13), MAC (#13), UNIX (#10).

  2.
  v 1.8 1999/10/25 10:59:15 alexm Exp
  hacked by alexm@hsys.msk.ru for Delphi 2.0 compatibility
  - yyinput and yyoutput can now be TStream's according to $DEFINE's

  1.
  Standard Lex library unit for TP Lex Version 3.0. 2-11-91 AG

*)

//Define VER140PLUS for Delphi 6 and next
{$IFDEF VER140} {$DEFINE VER140PLUS} {$ENDIF}
{$IFDEF VER150} {$DEFINE VER140PLUS} {$ENDIF}
{$IFDEF VER160} {$DEFINE VER140PLUS} {$ENDIF}
//Do not define VER140PLUS for Kylix 1 ; Define VER140PLUS for Kylix 2 and next.
{$DEFINE LEX_USES_STREAMS}
{$DEFINE LEX_USES_HUGE_STRINGS}

interface
{$IFDEF LEX_USES_STREAMS} uses Classes; {$ENDIF}

(* The Lex library unit supplies a collection of variables and routines
   needed by the lexical analyzer routine yylex and application programs
   using Lex-generated lexical analyzers. It also provides access to the
   input/output streams used by the lexical analyzer and the text of the
   matched string, and provides some utility functions which may be used
   in actions.

   This `standard' version of the LexLib unit is used to implement lexical
   analyzers which read from and write to MS-DOS files (using standard input
   and output, by default). It is suitable for many standard applications
   for lexical analyzers, such as text conversion tools or compilers.

   However, you may create your own version of the LexLib unit, tailored to
   your target applications. In particular, you may wish to provide another
   set of I/O functions, e.g., if you want to read from or write to memory
   instead to files, or want to use different file types. *)

(* Variables:

   The variable yytext contains the current match, yyleng its length.
   The variable yyline contains the current input line, and yylineno and
   yycolno denote the current input position (line, column). These values
   are often used in giving error diagnostics (however, they will only be
   meaningful if there is no rescanning across line ends).

   The variables yyinput and yyoutput are the text files which are used
   by the lexical analyzer. By default, they are assigned to standard
   input and output, but you may change these assignments to fit your
   target application (use the Turbo Pascal standard routines assign,
   reset, and rewrite for this purpose). *)

var
  {$IFDEF LEX_USES_STREAMS}
  yyinput, yyoutput : TStream;     (* input and output streams *)
  {$ELSE}
  yyinput, yyoutput : Text;        (* input and output file *)
  {$ENDIF}
  yyline            : String;      (* current input line *)
  yylineno, yycolno : Integer;     (* current input position *)
  yytext            : String;      (* matched text (should be considered r/o) *)

{$IFDEF LEX_USES_HUGE_STRINGS}
function yyleng():Integer;
{$ELSE}
yyleng : Byte absolute yytext;  // length of matched text
{$ENDIF}

(* I/O routines:

   The following routines get_char, unget_char and put_char are used to
   implement access to the input and output files. Since \n (newline) for
   Lex means line end, the I/O routines have to translate MS-DOS line ends
   (carriage-return/line-feed) into newline characters and vice versa. Input
   is buffered to allow rescanning text (via unput_char).

   The input buffer holds the text of the line to be scanned. When the input
   buffer empties, a new line is obtained from the input stream. Characters
   can be returned to the input buffer by calls to unget_char. At end-of-
   file a null character is returned.

   The input routines also keep track of the input position and set the
   yyline, yylineno, yycolno variables accordingly.

   Since the rest of the Lex library only depends on these three routines
   (there are no direct references to the yyinput and yyoutput files or
   to the input buffer), you can easily replace get_char, unget_char and
   put_char by another suitable set of routines, e.g. if you want to read
   from/write to memory, etc. *)

// obtain one character from the input file (null character at end-of-file)
function get_char():Char;

//return one character to the input file to be reread in subsequent calls to get_char
procedure unget_char(c:Char);

procedure put_char(c:Char); // write one character to the output file


// Utility routines: -------------------------------------------------------------------------
procedure echo(); //echoes the current match to the output stream
procedure yymore(); //append the next match to the current one
procedure yyless(n:Integer); //truncate yytext to size n and return the remaining characters to the input stream

procedure reject(); //reject the current match and execute the next one
(* reject does not actually cause the input to be rescanned; instead,
  internal state information is used to find the next match. Hence
  you should not try to modify the input stream or the yytext variable
  when rejecting a match. *)

procedure return(n:Integer);
procedure returnc(c:Char); //sets the return value of yylex

procedure start(state:Integer);
(* puts the lexical analyzer in the given start state; state=0 denotes
  the default start state, other values are user-defined *)

function yywrap():Boolean;
(* The yywrap function is called by yylex at end-of-file (unless you have
  specified a rule matching end-of-file). You may redefine this routine
  in your Lex program to do application-dependent processing at end of
  file. In particular, yywrap may arrange for more input and return false
  in which case the yylex routine resumes lexical analysis.

  The default yywrap routine supplied here closes input and output files
  and returns true (causing yylex to terminate). *)


(* The following are the internal data structures and routines ---------------------------------
  used by the lexical analyzer routine yylex; they should not be used directly. *)
var

yystate    : Integer; (* current state of lexical analyzer *)
yyactchar  : Char;    (* current character *)
yylastchar : Char;    (* last matched character (#0 if none) *)
yyrule     : Integer; (* matched rule *)
yyreject   : Boolean; (* current match rejected? *)
yydone     : Boolean; (* yylex return value set? *)
yyretval   : Integer; (* yylex return value *)

procedure yynew(); //starts next match; initializes state information of the lexical analyzer
procedure yyscan(); // gets next character from the input stream and updates yytext and yyactchar accordingly
procedure yymark(n:Integer); // marks position for rule no. n
procedure yymatch(n:Integer); // declares a match for rule number n

function yyfind(var n:Integer):Boolean;
(* finds the last match and the corresponding marked position and adjusts
  the matched string accordingly; returns:
    - true if a rule has been matched, false otherwise
    - n: the number of the matched rule *)

function yydefault():Boolean; //executes the default action (copy character); returns true unless at end-of-file
procedure yyclear(); // reinitializes state information after lexical analysis has been finished

type TCaracteres = set of Char;     
 //Return index of first caracter in "text" found
function strPosChar(const text:String; const caracteres:TCaracteres):Integer;
{$IFNDEF VER140PLUS}function ReverseString(const AText: string): string;{$ENDIF} //in StrUtils since Delphi6



implementation
{$IFDEF LEX_USES_STREAMS}uses SysUtils {$IFDEF VER140PLUS}, StrUtils{$ENDIF};
{$ELSE}
{$IFDEF VER140PLUS}uses StrUtils;{$ENDIF}
{$ENDIF}

{$IFDEF LEX_USES_HUGE_STRINGS}
function yyleng():Integer;
begin
  Result := Length(yytext);
end; //yyleng
{$ENDIF}

procedure fatal(const msg:String); //writes a fatal error message and halts program
begin
  writeln('LexLib: ', msg);
  halt(1);
end; //fatal


// I/O routines: --------------------------------------------

const nl = #10;  (* newline character *)

var
  bufptr : Integer;
  buf    : String;

{$IFDEF LEX_USES_STREAMS} { lex will use TStream's }
function GetLineFromStream(AStream : TStream) : String;
const BUFSIZ = 256 ;
var
  eol_flag : Boolean;
  arr : array[0..BUFSIZ - 1] of Char;
  charsRead, eol_pos : Integer;
  fin1,fin2 : Char;
  stringTmp:String;
begin
  Result := '';
  eol_flag := False;
  repeat //tant qu'on n'a pas une ligne complte
    charsRead := AStream.Read(arr, BUFSIZ); //lit un buffer
    if charsRead = 0 then eol_flag := True //fin de fichier
    else
    begin
      SetString(stringTmp, arr, charsRead);
      eol_pos := strPosChar(stringTmp,[#10,#13]); //repre le 1er signe de saut de ligne
      if eol_pos > 0 then //si un signe de saut de ligne a t repr
      begin
        eol_flag:=True;
        Result:=Result + Copy(stringTmp,1,Pred(eol_pos)); //on retourne la ligne lue sans CR ni LF
        fin1 := stringTmp[eol_pos]; //1er caractre de ligne
        AStream.Seek(- charsRead + eol_pos, soFromCurrent); //on revient aprs le 1er signe de ligne
        if fin1=#13 then //PC ou MAC
        begin
          if eol_pos < Length(stringTmp) then //on avait dj lu la suite
          begin
            fin2:=stringTmp[Succ(eol_pos)]; //l'ventuel 2me signe de ligne
            if fin2=#10 then AStream.Seek(1, soFromCurrent); //(PC) il faut sauter le 2me signe de ligne
          end
          else if AStream.Position < AStream.Size then //on tait en fin de buffer, on lit le caractre suivant si on n'est pas en fin de fichier 
          begin
            AStream.Read(arr,1);
            fin2:=arr[0];
            if fin2<>#10 then AStream.Seek(-1, soFromCurrent); //(MAC) il n'y avais pas de 2me signe de ligne, on revient en arrire
          end;
        end //si on a #13#10 on a avanc de 1 caractre sinon pas boug
        else //if fin1=#10 then //Unix ou PC invers
        begin
          if eol_pos < Length(stringTmp) then //on avait dj lu la suite
          begin
            fin2:=stringTmp[Succ(eol_pos)]; //l'ventuel 2me signe de ligne
            if fin2=#13 then AStream.Seek(1, soFromCurrent); //(PC invers) il faut sauter le 2me signe de ligne
          end
          else if AStream.Position < AStream.Size then //on tait en fin de buffer, on lit le caractre suivant si on n'est pas en fin de fichier
          begin
            AStream.Read(arr,1);
            fin2:=arr[0];
            if fin2<>#13 then AStream.Seek(-1, soFromCurrent); //(Unix) il n'y avais pas de 2me signe de ligne, on revient en arrire
          end;
        end; //si on a #10#13 on a avanc de 1 caractre sinon pas boug
      end
      else if AStream.Position >= Pred(AStream.Size) then //on est en fin de fichier
      begin
        eol_flag:=True;            //on fait comme si on tait en fin de ligne
        Result:=Result + stringTmp;
      end
      else Result:=Result + stringTmp; //aucun signe de saut de ligne n'a t repr, on continue  lire
    end;
  until eol_flag;
end; //GetLineFromStream
{$ENDIF}

function get_char():Char;
begin
  {$IFDEF LEX_USES_STREAMS}
  if (bufptr=0) and (yyinput.position < yyinput.size) then
  {$ELSE}
  if (bufptr=0) and not eof(yyinput) then
  {$ENDIF}
  begin
    {$IFDEF LEX_USES_STREAMS}
    yyline := GetLineFromStream(yyinput);
    {$ELSE}
    readln(yyinput, yyline);
    {$ENDIF}
    Inc(yylineno);
    yycolno := 1;
    SetLength(buf, Length(yyline) + 1);
    buf:=nl + ReverseString(yyline);
    Inc(bufptr, length(yyline)+1);
  end;
  if bufptr>0 then
  begin
    Result := buf[bufptr];
    Dec(bufptr);
    Inc(yycolno);
  end
  else Result := #0;
end; //get_char

procedure unget_char(c:Char);
begin
  if bufptr=Length(buf) then
    SetLength(buf, Length(buf) + 1);
  Inc(bufptr);
  Dec(yycolno);
  buf[bufptr] := c;
end; //unget_char

{$IFDEF LEX_USES_STREAMS}
procedure put_char(c:Char);
const cc : array[0..1] of char = #13#10;
begin
  if c=#0 then
    { ignore }
  else begin
    if yyoutput = nil then
      raise Exception.Create('LexLib: yyoutput = nil, c = "' + c + '"');
    if c=nl then
      yyoutput.Write(cc, 2)
    else
      yyoutput.Write(c, 1);
  end;
end; //put_char
{$ELSE}
procedure put_char(c:Char);
begin
  if c=#0 then
    { ignore }
  else if c=nl then
    writeln(yyoutput)
  else
    write(yyoutput, c)
end; //put_char
{$ENDIF}

(* Variables:

   Some state information is maintained to keep track with calls to yymore,
   yyless, reject, start and yymatch/yymark, and to initialize state
   information used by the lexical analyzer.
   - yystext: contains the initial contents of the yytext variable; this
     will be the empty string, unless yymore is called which sets yystext
     to the current yytext
   - yysstate: start state of lexical analyzer (set to 0 during
     initialization, and modified in calls to the start routine)
   - yylstate: line state information (1 if at beginning of line, 0
     otherwise)
   - yystack: stack containing matched rules; yymatches contains the number of
     matches
   - yypos: for each rule the last marked position (yymark); zeroed when rule
     has already been considered
   - yysleng: copy of the original yyleng used to restore state information
     when reject is used *)

const
  max_matches = 1024;
  max_rules   = 256;

var
  yystext            : String;
  yysstate, yylstate : Integer;
  yymatches          : Integer;
  yystack            : array [1..max_matches] of Integer;
  yypos              : array [1..max_rules] of Integer;
  yysleng            : Byte;

// Utilities: -------------------------------------------------

procedure echo();
var i : Integer;
begin
  for i := 1 to yyleng do
    put_char(yytext[i])
end; //echo

procedure yymore();
begin
  yystext := yytext;
end; //yymore

procedure yyless(n:Integer);
var i:Integer;
begin
  for i := yyleng downto n+1 do
    unget_char(yytext[i]);
  {$IFDEF WIN16}
  yytext[0] := n;
  {$ELSE}
  SetLength(yytext, n);
  {$ENDIF}
end; //yyless

procedure reject();
var i : Integer;
begin
  yyreject := True;
  for i := yyleng+1 to yysleng do
    yytext := yytext+get_char;
  Dec(yymatches);
end; //reject

procedure return(n:Integer);
begin
  yyretval := n;
  yydone := True;
end; //return

procedure returnc(c:Char);
begin
  yyretval := Ord(c);
  yydone := True;
end; //returnc

procedure start ( state : Integer );
begin
  yysstate := state;
end; //start

function yywrap():Boolean;
begin
  {$IFNDEF LEX_USES_STREAMS}
  close(yyinput);
  close(yyoutput);
  {$ENDIF}
  Result := True;
end; //yywrap


// Internal routines: ------------------------------------------

procedure yynew();
begin
  if yylastchar<>#0 then
    if yylastchar=nl then yylstate := 1
    else yylstate := 0;
  yystate := yysstate+yylstate;
  yytext  := yystext;
  yystext := '';
  yymatches := 0;
  yydone := false;
end; //yynew

procedure yyscan();
begin
  if yyleng=255 then fatal('yytext overflow');
  yyactchar := get_char;
  yytext := yytext + yyactchar;
end; //yyscan

procedure yymark(n:Integer);
begin
  if n>max_rules then fatal('too many rules');
  yypos[n] := yyleng;
end; //yymark

procedure yymatch(n:Integer);
begin
  Inc(yymatches);
  if yymatches>max_matches then fatal('match stack overflow');
  yystack[yymatches] := n;
end; //yymatch

function yyfind(var n:Integer):Boolean;
begin
  yyreject := False;
  while (yymatches>0) and (yypos[yystack[yymatches]]=0) do
    Dec(yymatches);
  if yymatches>0 then
  begin
    yysleng := yyleng;
    n       := yystack[yymatches];
    yyless(yypos[n]);
    yypos[n] := 0;
    if yyleng>0 then yylastchar := yytext[yyleng]
    else yylastchar := #0;
    Result := True;
  end
  else
  begin
    yyless(0);
    yylastchar := #0;
    Result := False;
  end
end; //yyfind

function yydefault():Boolean;
begin
  yyreject := False;
  yyactchar := get_char();
  if yyactchar<>#0 then
  begin
    put_char(yyactchar);
    Result := True;
  end
  else
  begin
    yylstate := 1;
    Result := False;
  end;
  yylastchar := yyactchar;
end; //yydefault

procedure yyclear();
begin
  bufptr := 0;
  buf := '';
  yysstate := 0;
  yylstate := 1;
  yylastchar := #0;
  yytext := '';
  yystext := '';
end;  //yyclear

function strPosChar(const text:String; const caracteres:TCaracteres):Integer;
var cpt:Integer; found:Boolean;
begin
  cpt:=1;
  repeat
    found := text[cpt] in caracteres;
    Inc(cpt);
  until found or (cpt>Length(text));
  if found then Result:=Pred(cpt)
  else Result:=0;
end;

{$IFNDEF VER140PLUS}
function ReverseString(const AText: string): string;
var
  I: Integer;
  P: PChar;
begin
  SetLength(Result, Length(AText));
  P := PChar(Result);
  for I := Length(AText) downto 1 do
  begin
    P^ := AText[I];
    Inc(P);
  end;
end;
{$ENDIF}

begin
{$IFDEF LEX_USES_STREAMS}
  yyinput := nil;
  yyoutput := nil;
{$ELSE}
  assign(yyinput, '');
  assign(yyoutput, '');
  reset(yyinput);
  rewrite(yyoutput);
{$ENDIF}
  yylineno := 0;
  yyclear();
end. //LexLib

